#
# This file is part of the magnum.np distribution
# (https://gitlab.com/magnum.np/magnum.np).
# Copyright (c) 2023 magnum.np team.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

from magnumnp.common import logging, timedmethod, constants
import torch

__all__ = ["OhmSolver"]

class OhmSolver(object):
    def __init__(self, dirichlet_bc_nodes):
        self._dirichlet_bc_nodes = dirichlet_bc_nodes

    def u(self, state, **kwargs):
        dx = state.mesh.dx
        sigma = state.material["sigma"]
        rhs = state.Constant(0.)
        u0 = state.u.clone()

        def _M(u):
            res = torch.zeros_like(u)

            # homogeneous Neumann conditions for now
            if u.shape[0] > 2:
                u[ 0,:,:] = u[ 1,:,:]
                u[-1,:,:] = u[-2,:,:]
            if u.shape[1] > 2:
                u[:, 0,:] = u[:, 1,:]
                u[:,-1,:] = u[:,-2,:]
            if u.shape[2] > 2:
                u[:,:, 0] = u[:,:, 1]
                u[:,:,-1] = u[:,:,-2]

            # set Dirichlet conditions
            u[self._dirichlet_bc_nodes] = u0[self._dirichlet_bc_nodes]

            # assemble laplace
            # x
            sigma_avg = 2. * sigma[1:,:,:] * sigma[:-1,:,:] / (sigma[1:,:,:] + sigma[:-1,:,:])
            sigma_avg = sigma_avg.nan_to_num(posinf=0, neginf=0)
            du = u[1:,:,:] - u[:-1,:,:]
            du *= sigma_avg / dx[0]
            res[1:-1,:,:] += (du[1:,:,:] - du[:-1,:,:]) / dx[0]

            # y
            sigma_avg = 2. * sigma[:,1:,:] * sigma[:,:-1,:] / (sigma[:,1:,:] + sigma[:,:-1,:])
            sigma_avg = sigma_avg.nan_to_num(posinf=0, neginf=0)
            du = u[:,1:,:] - u[:,:-1,:]
            du *= sigma_avg / dx[1]
            res[:,1:-1,:] += (du[:,1:,:] - du[:,:-1,:]) / dx[1]

            # z
            sigma_avg = 2. * sigma[:,:,1:] * sigma[:,:,:-1] / (sigma[:,:,1:] + sigma[:,:,:-1])
            sigma_avg = sigma_avg.nan_to_num(posinf=0, neginf=0)
            du = u[:,:,1:] - u[:,:,:-1]
            du *= sigma_avg / dx[2]
            res[:,:,1:-1] += (du[:,:,1:] - du[:,:,:-1]) / dx[2]

            return res

        state.u = conjugate_gradient(_M, state.u, rhs, **kwargs)
        return state.u

    def j(self, state, **kwargs):
        sigma = state.material["sigma"]
        u = self.u(state, **kwargs).squeeze(-1)
        grad = [torch.gradient(u, dim=i)[0] if state.mesh.n[i] > 2 else torch.zeros_like(u) for i in range(3)]
        return -sigma * torch.stack(grad, dim=-1)

def conjugate_gradient(A, x, b, tol=1e-6, max_iter=1000):
    r = b - A(x)
    p = r.clone()
    rsold = (r*r).sum()

    for i in range(max_iter):
        Ap = A(p)
        pAp = (p*Ap).sum()
        alpha = rsold / pAp
        x = x + alpha * p
        r = r - alpha * Ap
        rsnew = (r*r).sum()
        norm = torch.norm(r)
        if norm < tol:
            break
        beta = rsnew / rsold
        p = r + beta * p
        rsold = rsnew
        if i % 100 ==0:
            logging.info_blue("[OhmSolver] it= %d, norm= %.5e" % (i, norm))
    return x



#
# This file is part of the magnum.np distribution
# (https://gitlab.com/magnum.np/magnum.np).
# Copyright (c) 2023 magnum.np team.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

import torch
import numpy as np

__all__ = ["LogDt", "LogMovingAverage", "LogCumSum"]

class LogDt():
    ''' Discrete difference '''
    def __init__(self, func):
        self._func = func

    def __call__(self, state):
        try:
            f0 = self._f
            t0 = self._t
        except:
            f0 = 0
            t0 = 0
        self._f = self._func(state)
        self._t = state.t
        return (self._f - f0) / (self._t - t0)

class LogMovingAverage():
    def __init__(self, func, N = 10):
        self._func = func
        self._vals = np.zeros(N)
        self._fill = np.zeros(N)

    def __call__(self, state):
        self._vals = np.roll(self._vals, -1)
        self._fill = np.roll(self._fill, -1)

        self._vals[-1] = self._func(state)
        self._fill[-1] = 1
        return self._vals.sum() / self._fill.sum()

class LogCumSum():
    def __init__(self, func):
        self._func = func
        self._val = 0.

    def __call__(self, state):
        self._val += self._func(state)
        return self._val



#
# This file is part of the magnum.np distribution
# (https://gitlab.com/magnum.np/magnum.np).
# Copyright (c) 2023 magnum.np team.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

from magnumnp.common import logging, timedmethod, constants, complex_dtype
from .field_terms import FieldTerm
import numpy as np
import torch
import torch.fft
from torch import asinh, atan, sqrt, log, abs
import os
from time import time

__all__ = ["OerstedField"]

def g(x, y, z):
    R = sqrt(x**2 + y**2 + z**2)

    res = (3.*x**2 + 3.*y**2 - 2.*z**2)*z*R/24.
    res += np.pi*z/4.*abs(x*y*z)
    res += ((x**4 - 6.*x**2*y**2 + y**4)/24. * log(z+R)).nan_to_num(posinf=0, neginf=0)
    res += (x*y/6. * (y**2 - 3.*z**2) * atan(x*z/(y*R))).nan_to_num(posinf=0, neginf=0)
    res += (x*y/6. * (x**2 - 3.*z**2) * atan(y*z/(x*R))).nan_to_num(posinf=0, neginf=0)
    res += (z/6. * x * (z**2 - 3.*y**2) * log(x+R)).nan_to_num(posinf=0, neginf=0)
    res += (z/6. * y * (z**2 - 3.*x**2) * log(y+R)).nan_to_num(posinf=0, neginf=0)

    return res

def G1(x, y, z, dz):
    return - 1.*g(x, y, z+dz) \
           + 2.*g(x, y, z   ) \
           - 1.*g(x, y, z-dz)

def G0(x, y, z, dy, dz):
    return - 1.*G1(x, y+dy, z, dz) \
           + 2.*G1(x, y   , z, dz) \
           - 1.*G1(x, y-dy, z, dz)

def krueger_g(x, y, z, dx, dy, dz):
    ret = - 1.*G0(x+dx, y, z, dy, dz) \
          + 2.*G0(x   , y, z, dy, dz) \
          - 1.*G0(x-dx, y, z, dy, dz)
    return ret / (4.*np.pi*dx*dy*dz)


def dipole_g(x, y, z, dx, dy, dz):
    R = sqrt(x**2 + y**2 + z**2)
    res = -z/R**3
    res[0,0,0] = 0.
    return res * dx*dy*dz / (4.*np.pi)

def oersted_g(x, y, z, dx, dy, dz, p):
    res = dipole_g(x, y, z, dx, dy, dz)
    near = (x**2 + y**2 + z**2) / (dx**2 + dy**2 + dz**2) < p**2
    res[near] = krueger_g(x[near], y[near], z[near], dx, dy, dz)
    return res


class OerstedField(FieldTerm):
    r"""
    The Oersted field created by some current density :math:`\vec{j}` can be calculated by means of the Biot-Savart law

    .. math::

        \vec{h}^\text{oersted}(\vec{x}) = \frac{1}{4 \pi} \int \vec{j}(\vec{x}') \times \frac{\vec{x}-\vec{x}'}{\vert \vec{x}-\vec{x}'\vert^3} \, d\vec{x}'.

    The occuring equations [krueger] look very similar to those of the demag field, and the occuring convolution can be efficiently calculated by means of an FFT method.

    :param p: number of next neighbors for near field via Krueger's equations (default = 20)
    :type p: int, optional
    """
    def __init__(self, p = 20, cache_dir = None):
        self._p = p
        self._cache_dir = cache_dir

    def _init_K_component(self, state, perm, func):
        # dipole far-field
        dx = np.array(state.mesh.dx)

        shape = [1 if n==1 else 2*n for n in state.mesh.n]
        ij = [torch.fft.fftfreq(n,1/n) for n in shape] # local indices
        ij = torch.meshgrid(*ij,indexing='ij')
        x, y, z = [ij[ind]*dx[ind] for ind in perm]
        dx = [dx[ind] for ind in perm]

        Kc = func(x, y, z, *dx, self._p) # TODO: handle PBCs and non-equidistant grids

        dim = [i for i in range(3) if state.mesh.n[i] > 1]
        if len(dim) > 0:
            Kc = torch.fft.rfftn(Kc, dim = dim)

        return Kc #.real.clone() ## Oersted Kernel is not symmetric

    def _init_K(self, state):
        if state.mesh.pbc[0] != 0 or state.mesh.pbc[1] != 0 or state.mesh.pbc[2] != 0:
            logging.warning(f"[OERSTED]: PBCs are not used by OerstedField! (mesh.pbc = %s)" % str(state.mesh.pbc))

        name = "/K_%s.pt" % str(state.mesh).replace(" ","")
        if self._cache_dir != None and os.path.isfile(self._cache_dir + name):
            [Kxy, Kyz, Kxz] = torch.load(self._cache_dir + name, map_location=state.device)
            logging.info("[OERSTED]: Use cached Oersted kernel from '%s'" % (self._cache_dir + name))
        else:
            dtype = torch.get_default_dtype()
            torch.set_default_dtype(torch.float64) # always use double precision
            time_kernel = time()

            Kxy = self._init_K_component(state, [0,1,2], oersted_g).to(dtype=complex_dtype[dtype])
            Kyz = self._init_K_component(state, [1,2,0], oersted_g).to(dtype=complex_dtype[dtype])
            Kxz = self._init_K_component(state, [2,0,1], oersted_g).to(dtype=complex_dtype[dtype])

            logging.info(f"[OERSTED]: Time calculation of oersted kernel = {time() - time_kernel} s")
            torch.set_default_dtype(dtype) # restore dtype

            # cache Oersted tensor
            if self._cache_dir != None:
                if not os.path.isdir(self._cache_dir):
                    os.makedirs(self._cache_dir)
                torch.save([Kxy, Kyz, Kxz], self._cache_dir + name)
                logging.info("[OERSTED]: Save Oersted kernel to '%s'" % (self._cache_dir + name))

        return [[  0., -Kxy, +Kxz],
                [+Kxy,   0., -Kyz],
                [-Kxz, +Kyz,   0.]]

    @timedmethod
    def h(self, state):
        if not hasattr(self, "_K"):
            self._K = self._init_K(state)

        hx = torch.zeros_like(self._K[0][1])
        hy = torch.zeros_like(self._K[0][1])
        hz = torch.zeros_like(self._K[0][1])

        j = state.j # state calls j(state) if j is a function

        for ax in range(3):
            j_pad_fft1D = torch.fft.rfftn(j[:,:,:,ax], dim = [i for i in range(3) if state.mesh.n[i] > 1], s = [2*state.mesh.n[i] for i in range(3) if state.mesh.n[i] > 1])

            hx += self._K[0][ax] * j_pad_fft1D
            hy += self._K[1][ax] * j_pad_fft1D
            hz += self._K[2][ax] * j_pad_fft1D

        hx = torch.fft.irfftn(hx, dim = [i for i in range(3) if state.mesh.n[i] > 1])
        hy = torch.fft.irfftn(hy, dim = [i for i in range(3) if state.mesh.n[i] > 1])
        hz = torch.fft.irfftn(hz, dim = [i for i in range(3) if state.mesh.n[i] > 1])

        return torch.stack([hx[:state.mesh.n[0],:state.mesh.n[1],:state.mesh.n[2]],
                            hy[:state.mesh.n[0],:state.mesh.n[1],:state.mesh.n[2]],
                            hz[:state.mesh.n[0],:state.mesh.n[1],:state.mesh.n[2]]], dim=3)



#
# This file is part of the magnum.np distribution
# (https://gitlab.com/magnum.np/magnum.np).
# Copyright (c) 2023 magnum.np team.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

import os
import torch
from magnumnp.common import logging, read_vti
from xml.etree import cElementTree
from xml.dom import minidom

__all__ = ["FieldLogger"]

class FieldLogger(object):
    def __init__(self, filename, fields, every = 1, scale = 1.):
        """
        Logger class for fields

        *Arguments*
            filename (:class:`str`)
                The name of the log file
            fields ([:class:`str` | :class:`function`])
                The columns to be written to the log file
            every (:class:`int`)
                Write row to log file every nth call
            scale (:class:`float`)
                Scale factor for dimentions (e.g. 1e9 for nm-units)

        *Example*
            .. code-block:: python

                # provide key strings with are available in state
                logger = FieldLogger('data/m.pvd', ['m', demag.h])

                # Actually log fields
                state = State(mesh)
                logger << state
        """
        # create directory if not existent
        if not os.path.dirname(filename) == '' and \
             not os.path.exists(os.path.dirname(filename)):
            try:
                os.makedirs(os.path.dirname(filename))
            except OSError as exc: # Guard against race condition
                if exc.errno != errno.EEXIST:
                    raise

        filename, ext = os.path.splitext(filename)
        if ext != ".pvd":
            raise NameError("Only .pvd extention allowed")
        self._filename = filename
        self._every = every
        self._scale = scale
        if isinstance(fields, str):
            fields = [fields]
        self._fields = fields
        self._i = 0
        self._i_start = 0
        self._xmlroot = cElementTree.Element("VTKFile", type="Collection", version="0.1", byte_order="LittleEndian")
        cElementTree.SubElement(self._xmlroot, "Collection")

    def log(self, state):
        self._i += 1
        if ((self._i-1) % self._every > 0):
            return
        if (self._i <= self._i_start):
            return

        values = {}
        n_unnamed = 0
        for field in self._fields:
            if isinstance(field, str):
                name = field
                value = getattr(state, field)
            elif hasattr(field, '__call__'):
                if hasattr(field, '__self__') and hasattr(field.__self__, '__name__'):
                    name = field.__self__.__class__.__name__ + "." + field.__name__
                elif hasattr(field, '__name__'):
                    name = field.__name__
                else:
                    name = 'unnamed'

                if name == 'unnamed' or name == '<lambda>':
                    name = 'unnamed%04d' % n_unnamed
                    n_unnamed += 1

                value = field(state)
            elif isinstance(field, tuple) or isinstance(field, list):
                name = field[0]
                value = field[1]
            elif isinstance(value, torch.Tensor):
                name = 'unnamed'
                value = field
            else:
                raise RuntimeError('[FieldLogger] Column type not supported!')

            if hasattr(value, '__call__'):
                if name == 'unnamed':
                    try:
                        name = value.__self__.__class__.__name__ + "." + value.__name__
                    except:
                        pass
                value = value(state)
            values[name] = value

        filename = "%s_%04d" % (self._filename, self._i // self._every)
        if state.mesh.is_equidistant:
            filename += ".vti"
        else:
            filename += ".vtr"
        state.write_vtk(values, filename, scale = self._scale)

        cElementTree.SubElement(self._xmlroot[0], "DataSet", timestep=str(float(state.t)), file=os.path.basename(filename))
        with open(self._filename + ".pvd", 'w') as fd:
            fd.write(minidom.parseString(" ".join(cElementTree.tostring(self._xmlroot).decode().replace("\n","").split()).replace("> <", "><")).toprettyxml(indent="  "))
            fd.flush()

    def __lshift__(self, state):
        self.log(state)

    def reset(self):
        self._i = 0

    def resumable_step(self):
        try:
            xml = cElementTree.parse(self._filename + ".pvd").getroot()
            return len(list(xml.find('Collection'))) * self._every
        except IOError:
            return 0

    def last_recorded_step(self):
        """
        Returns the number of the last step logged and None if no
        step was yet logged.

        *Returns*
            :class:`int`
                Number of the last step recorded
        """
        result = (self.resumable_step() // self._every - 1) * self._every
        if result < 0:
            return None
        else:
            return result

    def step_data(self, i, field = None):
        """
        Returns field and time to a given step number.

        *Arguments*
            i (:class:`int`)
                The step number
            field (:class:`str`)
                The field to be read

        *Returns*
            (:class:`dolfin.Function`, :class:`float`)
                The field of step i and the corresponding time
        """
        if i % self._every > 0:
            raise Exception()

        xml = cElementTree.parse(self._filename + ".pvd").getroot()
        item = list(xml.find('Collection'))[i // self._every]
        mesh, data = read_vti(os.path.join(os.path.dirname(self._filename), item.attrib['file']))

        return data[field], float(item.attrib['timestep'])

    def resume(self, i):
        """
        Try to resume existing log file from log step i. The log file
        is truncated accordingly.

        *Arguments*
            i (:class:`int`)
                The log step to resume from
        """
        self._i = i
        self._i_start = self.last_recorded_step() + 1
        self._xmlroot = cElementTree.parse(self._filename + ".pvd").getroot()



#
# This file is part of the magnum.np distribution
# (https://gitlab.com/magnum.np/magnum.np).
# Copyright (c) 2023 magnum.np team.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

import torch
from magnumnp.common import logging, Material
from magnumnp.common.io import write_vti, write_vtr

__all__ = ["complex_dtype", "normalize", "randM", "Expression"]


complex_dtype = {
    torch.float: torch.complex,
    torch.float32: torch.complex64,
    torch.float64: torch.complex128
    }


def normalize(data):
    r"""
    Helper function to normalize vectorial data inplace
    """
    data /= torch.linalg.norm(data, dim = -1, keepdim = True)
    data[...] = torch.nan_to_num(data, posinf=0, neginf=0)
    return data

def randM(data):
    r"""
    Helper function to generate uniform distibution on the unit-sphere
    """
    if data.shape[-1] != 3:
        raise ValueError("Input tensor's last dimension needs to be 3 (shape='%s')" % str(data.shape))

    theta = 2.*torch.pi*torch.rand(data.shape[:-1])
    phi = torch.acos(2.*torch.rand(data.shape[:-1])-1.)

    data[...,0] = torch.sin(phi) * torch.cos(theta)
    data[...,1] = torch.sin(phi) * torch.sin(theta)
    data[...,2] = torch.cos(phi)
    return data


def Expression(comps):
    r"""
    Helper function to create scalar or vector fields by stacking the corresponding components.

    :param comps: list of components or single torch.Tensor
    :type comps: list, :class:`Tensor`

    :Examples:

    .. code::
        n  = (1, 1, 10)
        dx = (2e-9, 2e-9, 2e-9)
        mesh = Mesh(n, dx)
        x, y, z = mesh.SpatialCoordinate()
        
        Ms = Expression(x)            # create scalar field
        Ku_axis = Expression((x,y,z)) # create vector field
    """
    if isinstance(comps, torch.Tensor):
        comps = [comps]
    return torch.stack(comps, dim=-1)


def get_gpu_with_least_memory():
    if not torch.cuda.is_available():
        return -1

    import pynvml
    pynvml.nvmlInit()
    num_gpus = pynvml.nvmlDeviceGetCount()


    if num_gpus == 1:
        pynvml.nvmlShutdown()
        return 0
    
    else:
        gpu_memory = []
        for i in range(num_gpus):
            handle = pynvml.nvmlDeviceGetHandleByIndex(i)
            mem_info = pynvml.nvmlDeviceGetMemoryInfo(handle)
            gpu_memory.append(mem_info.used)

        pynvml.nvmlShutdown()
        return gpu_memory.index(min(gpu_memory))



#
# This file is part of the magnum.np distribution
# (https://gitlab.com/magnum.np/magnum.np).
# Copyright (c) 2023 magnum.np team.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

from magnumnp.common import constants, complex_dtype
from magnumnp.common.mesh import Mesh
from magnumnp.common.state import State
from magnumnp.field_terms import DemagField
import torch
from math import sin, cos, pi, sqrt
import numpy as np

__all__ = ["LTEM", "MFM", "to_discretisedfield"]

class LTEM(object): # TODO: document and improve interface
    def __init__(self, state, method = "beleggia", comp = 2, voltage = 300e3, kcx = 0.1, kcy = 0.1, theta = 0.):
        self._state = state
        self._mesh = state.mesh
        self._Ms = self._state.material["Ms"]
        self._M = self._state.m * self._state.material["Ms"]
        self._h = self._mesh.n[comp] * self._mesh.dx[comp]
        self._dim = [0, 1]
        self._volt = voltage
        self._kcx, self._kcy = kcx, kcy
        self._comp = comp
        self._theta = np.deg2rad(theta)
        self._method = method

    def _lambda_el(self):
        return 2.*pi*constants.hbar/sqrt(2 * constants.me * self._volt * constants.e * (1 + constants.e * self._volt / (2 * constants.me * constants.c**2)))

    def _k(self):
        dim = [ni for ni in self._mesh.n]
        dim[self._comp] = 1
        ij = [torch.fft.fftshift(torch.fft.fftfreq(dim[ind], self._mesh.dx[ind])) for ind in [0,1,2]]
        dk = [abs(ij[0][1]-ij[0][0]), abs(ij[1][1]-ij[1][0])]
        ij = torch.meshgrid(*ij, indexing = 'ij')
        k = torch.stack(ij, dim = -1).squeeze(dim = 2).to(dtype = complex_dtype[self._state.dtype])
        return k, dk

    def _Phim(self):
        m_int = torch.sum(self._M, dim = self._comp)*self._mesh.dx[self._comp]
        Mmn = torch.fft.fftn(m_int, dim = self._dim)

        k, dk = self._k()
        kx = k.real[:,:,0]
        ky = k.real[:,:,1]

        denom = (kx**2. +  ky**2.)/(kx**2. + ky**2. + dk[0]**2 * self._kcx**2. + self._kcy**2 * dk[1]**2. )**2.
        denom = denom.unsqueeze(-1)
        prefactor = 1j*constants.mu_0*constants.e / (2.*pi*constants.hbar)
        Fphim = denom * prefactor * torch.linalg.cross(Mmn, k)[:,:,2].unsqueeze(-1)
        Phim = torch.fft.ifftn(Fphim, dim = self._dim).real
        return Phim

    def Defocus(self, df, cs):
        k, dk = self._k()
        kx = k.real[:,:,0]
        ky = k.real[:,:,1]
        ks = kx**2. + ky**2.
        if self._method == "mansuripur":
            phim = self._Phim_Mansuripur()
        else:
            phim = self._Phim()

        cts = - df + 0.5 * self._lambda_el() ** 2. * cs * ks
        exp = torch.exp(pi * cts * 1j * ks * self._lambda_el()).unsqueeze(dim=-1)
        def_wf_cts = torch.fft.ifft2(torch.fft.fftshift(torch.fft.fft2(torch.exp(phim * 1j), dim = self._dim)) * exp, dim = self._dim)
        return (torch.conj(def_wf_cts) * def_wf_cts).real

    def MagneticPhaseShift(self):
        return torch.rad2deg(self._Phim())

    def InductionMap(self):
        if self._method == "mansuripur":
            phim = self._Phim_Mansuripur()
        else:
            phim = self._Phim()
        dphidx = torch.gradient(phim, dim = 0, spacing = self._mesh.dx[0])[0]
        dphidy = torch.gradient(phim, dim = 1, spacing = self._mesh.dx[1])[0]
        return constants.hbar/(constants.e*self._h)*torch.stack([-dphidy, dphidx], dim = -1)

    def _Phim_Mansuripur(self):
        m_int = torch.sum(self._M, dim = self._comp)*self._mesh.dx[self._comp]
        Mmn = torch.fft.fftn(m_int, dim = self._dim)
        p = torch.tensor([0., sin(self._theta), cos(self._theta)], dtype = complex_dtype[state.dtype]).unsqueeze(dim=0).unsqueeze(dim=0)
        ez = torch.tensor([0.,0.,1.], dtype = complex_dtype[state.dtype]).unsqueeze(dim=0).unsqueeze(dim=0)
        k, dk = self._k()
        ks = torch.linalg.norm(k, dim = -1, keepdim = True)# + (dk[0]**2 * self._kcx**2. + dk[1]**2 * self._kcy**2.)
        ks[0,0, :] = 1e-15
        kbool = (ks == 0)
        ks[kbool] = 1e-15

        pxpxM = torch.linalg.cross(p, torch.linalg.cross(p, Mmn))
        khat = k/ks
        Gp = 1./(torch.sum(p * khat, dim = -1, keepdim = True)**2. + p[:,:,2]**2.) * torch.sinc(self._h * ks * torch.sum(p * khat, dim = -1, keepdim = True)/ p[:,:,2])
        func = self._h / ks * Gp * (torch.sum(torch.linalg.cross(khat, ez) * pxpxM, dim = -1, keepdim = True))  * 1j
        phim = 2. * constants.e / constants.hbar * torch.fft.ifftn(func, dim = self._dim).real
        return phim


class MFM(object):
    def __init__(self, height = 100e-9, Q = 1, k = 1, mm_tip = 0., dm_tip = None):
        r""" Calculation of the phase shift of an MFM tip.

        The contrast in MFM images originates from the magnetic interaction between
        the magnetic tip of an oscillating cantilever and the samples stray field.
        As MFM is based on the stray field outside of a sample, an 'airbox' method
        is used. The saturation magnetisation has to be set to zero outside of the sample
        in which we wish to perform these MFM measurements.

        The magnetic cantilever is driven to oscillate near its resonant frequency
        when there is no stray field. In the presence of a stray magnetic field the
        phase shift of MFM cantilever is given by

        .. math::
            \Delta \phi = \frac{Q\mu_0}{k} \left( q \frac{\partial
                          {\bf H}_{sz}}{\partial z} + {\bf M}_t \cdot
                          \frac{\partial^2{\bf H}_{s}}{\partial z^2} \right),

        NOTE: this code is based on the MFM implementation of Ubermag
              (see also: https://ubermag.github.io/documentation/notebooks/mag2exp/Magnetic_Force_Microscopy.html)

        *Example*
          .. code:: python

            state = State(mesh)
            state.material = {"Ms": 8e5}
            state.material["Ms"][~magnetic] = 0. # zero Ms outside of sample

            x, y, z = meste.SpatialCoordinate()
            state.m = torch.stack([y, -x, 0*z], dim=-1)
            state.m.normalize()

            demag = DemagField()
            mfm = MFM(height=10e-9, mm_tip = 10e-9, dm_tip=20e-9)

            mfm_logger = FieldLogger("data/phi.pvd", [mfm.PhaseShift])
            mfm_logger << mfm.extended_state
        """
        self._prefactor = Q*constants.mu_0/k
        self._mm_tip = mm_tip
        self._dm_tip = dm_tip
        self._height = height
        self._xdemag = DemagField()

    def extend_demag(self, state):
        n = state.mesh.n
        if not hasattr(self, "extended_state"):
            dx = state.mesh.dx

            nz = int(n[2] + self._height/dx[2])
            n_new = [n[0], n[1], nz]

            mesh_new = Mesh(n_new, dx)
            self.extended_state = State(mesh_new)
            self.extended_state.material["Ms"] = 0.
            self.extended_state.m = self.extended_state.Constant([0.,0.,00.])

        # copy data
        self.extended_state.material["Ms"][:,:,:n[2]] = state.material["Ms"]
        self.extended_state.m[:,:,:n[2],:] = state.m
        return self._xdemag.h(self.extended_state)

    def PhaseShift(self, state, mm_tip = None, dm_tip = None):
        h_demag = self.extend_demag(state)
        spacing = state.mesh.dx[2]
        mm_tip = mm_tip or self._mm_tip
        dm_tip = dm_tip or self._dm_tip

        # monopole
        dHsdz = torch.gradient(h_demag, dim = [2], spacing = spacing)[0]
        phi = mm_tip * dHsdz[:,:,:,2]

        # dipole
        if dm_tip != None:
            d2Hsdz2 = torch.gradient(dHsdz, dim = [2], spacing = spacing)[0]
            phi += (dm_tip * d2Hsdz2).sum(dim=-1)

        return phi.unsqueeze(-1)


def to_discretisedfield(data):
    '''
    Utility function to convert vector fields to Ubermag's Field object.

    This allows using all sorts of post-processing features provided by Ubermag.
    The code has been contributed by @AbsoluteVacuum (see #33 in gitlab)
    '''
    from discretisedfield import Field
    from xarray import DataArray
    xarr = DataArray(
        data.detach().cpu().numpy(),
        dims=["x", "y", "z", "comp"],
        coords=dict(
            x=torch.arange(nx).numpy()*dx,
            y=torch.arange(ny).numpy()*dy,
            z=torch.arange(nz).numpy()*dz,
            comp=["x", "y", "z"],
        ),
        name="mag",
        attrs=dict(cell=mesh.dx, p1=mesh.origin, p2=[a*b+c for a,b,c in zip(mesh.n, mesh.dx, mesh.origin)])
    )
    return Field.from_xarray(xarr)



#
# This file is part of the magnum.np distribution
# (https://gitlab.com/magnum.np/magnum.np).
# Copyright (c) 2023 magnum.np team.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

import torch
from magnumnp.common import logging, Material
from magnumnp.common.io import write_vti, write_vtr
from magnumnp.common.utils import randM

__all__ = ["State"]


class State(object):
    def __init__(self, mesh):
        self.mesh = mesh

        self._material = Material(self)
        self._t = torch.tensor(0.)
        self._step = 0
        self._dt = 0.

        dtype_str = str(self.dtype).split('.')[1]
        logging.info_green("[State] running on device: %s (dtype = %s)" % (self.device, dtype_str))
        logging.info_green("[Mesh] %s" % mesh)

    # Time
    @property
    def t(self):
        return self._t

    @t.setter
    def t(self, value):
        if isinstance(value, (int, float)):
            self._t = torch.tensor(float(value))
        else:
            self._t = value

    # Material
    @property
    def material(self):
        return self._material

    @material.setter
    def material(self, values):
        if isinstance(values, dict):
            self._material = Material(self)
            for key, value in values.items():
                self._material[key] = value
        else:
            raise ValueError("Dictionary needs to be provided to set material")

    # Current density
    @property
    def j(self):
        return self._j(self)

    @j.setter
    def j(self, value):
        if callable(value):
            self._j = value
        else:
            self._j = lambda state: value

    # Temperature
    @property
    def T(self):
        return self._T(self)

    @T.setter
    def T(self, value):
        if callable(value):
            self._T = value
        else:
            self._T = lambda state: value

    @property
    def dtype(self):
        return self.mesh.dx_tensor[0].dtype

    @property
    def device(self):
        return self.mesh.dx_tensor[0].device

    def Constant(self, c, dtype = None, requires_grad = False):
        if not isinstance(c, torch.Tensor):
            c = torch.tensor(c, dtype = dtype, device = self.device)
        if c.dim() == 0 and c.dtype != torch.bool:
            c = c.reshape(1)
        x = torch.zeros(self.mesh.n + c.shape, dtype = dtype, device = self.device)
        x[...] = c
        if requires_grad == True:
            x.requires_grad = requires_grad
        return x

    def RandM(self):
        x = self.Constant([0.,0.,0.])
        randM(x)
        return x

    def SpatialCoordinate(self):
        logging.warning("State.SpatialCoordinate() is deprecated! Use mesh.SpatialCoordinate() instead!")
        return self.mesh.SpatialCoordinate()

    def convert_tensorfield(self, value):
        ''' convert arbitrary input to tensor-fields '''
        if not isinstance(value, torch.Tensor):
            value = torch.tensor(value, dtype=self.dtype)

        if len(value.shape) == 0: # convert dim=0 tensor into dim=1 tensor
            value = value.reshape(1)
        if len(value.shape) < 3: # expand homogeneous material to [nx,ny,nz,...] tensor-field
            shape = value.shape
            value = value.reshape((1,1,1) + tuple(shape))
            value = value.expand(self.mesh.n + tuple(shape))
            #value._expanded = True # annotate expanded tensor (clone will be before individual items are modified)
            value = value.clone()
        elif len(value.shape) == 3: # scalar-field should have dimension [nx,ny,nz,1]
            value = value.unsqueeze(-1)
        else: # otherwise assume the dimention is correct!
            pass
        return value

    def write_vtk(self, fields, filename, scale = 1.):
        filename = str(filename)
        if self.mesh.is_equidistant:
            if not filename.endswith(".vti"):
                filename += ".vti"
            write_vti(fields, filename, self, scale)
        else:
            if not filename.endswith(".vtr"):
                filename += ".vtr"
            write_vtr(fields, filename, self, scale)


    def avg(self, data, cell_volumes = None, dim=(0,1,2)):
        r"""
        Average over spatial dimensions of tensor fields.

        :param data: tensor field to average
        :type A: :class:`Tensor`
        :param dim: dimensions to average over
        :type dim: tuple, optional
        :param cell_volumes: volume of each cell (required only in case of non-equidistant meshes)
        :type cell_volumes: :class:`Tensor`, optional

        :Examples:

        .. code::
            Ms_avg = avg(state.material["Ms"])
            m_avg = avg(state.m)
        """
        if self.mesh.is_equidistant:
            if data.dim() <= 1: # e.g. [0,0,1]
                return data
            elif data.dim() == 2: # state.m[domain]
                return data.mean(dim=0)
            else:                 # [nx,ny,nz,...]
                return data.mean(dim=dim)
        else: # non-equidistant
            if cell_volumes == None:
                cell_volumes = self.mesh.cell_volumes
            if data.dim() <= 1: # e.g. [0,0,1]
                return data
            if data.dim() == 2: # state.m[domain]
                if data.shape[0] != cell_volumes.shape[0]:
                    raise ValueError("Data shape (%s) does not match cell_volumes shape (%s). When averaging over slices of non-equidistant tensors you have to provide a sliced version of state.mesh.cell_volumes!" % (str(data.shape), str(cell_volumes.shape)))
                return (data * cell_volumes).sum(dim=0) / cell_volumes.sum(dim=0)
            if data.dim() == 3: # [nx,ny,nz]
                if data.shape[:3] != cell_volumes.shape[:3]:
                    raise ValueError("Data shape (%s) does not match cell_volumes shape (%s). When averaging over slices of non-equidistant tensors you have to provide a sliced version of state.mesh.cell_volumes!" % (str(data.shape), str(cell_volumes.shape)))
                return (data * cell_volumes.squeeze(-1)).sum(dim=dim) / cell_volumes.sum()
            # [nx,ny,nz,...]
            return (data * cell_volumes).sum(dim=dim) / cell_volumes.sum()



#
# This file is part of the magnum.np distribution
# (https://gitlab.com/magnum.np/magnum.np).
# Copyright (c) 2023 magnum.np team.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

from magnumnp.common import timedmethod, constants
import torch
from .field_terms import LinearFieldTerm

__all__ = ["ExchangeField"]

class ExchangeField(LinearFieldTerm):
    r"""
    Exchange Field

    .. math::
        \vec{h}^\text{ex}_i = \frac{2}{\mu_0 \, M_{s,i}} \; \sum_{k=\pm x, \pm y,\pm z} \frac{2}{\Delta_k} \frac{A_{i+\vec{e}_k} \; A_i}{A_{i+\vec{e}_k} + A_i} \; \left( \vec{m}_{i+\vec{e}_k} - \vec{m}_i \right),

    with the vacuum permeability :math:`\mu_0`, the saturation magnetization :math:`M_s`, and the exchange constant :math:`A`. :math:`\Delta_k` and :math:`\vec{e}_k` represent the grid spacing and the unit vector in direction :math:`k`, respectively.

    :param A: Name of the material parameter for the exchange constant :math:`A`, defaults to "A"
    :type A: str, optional
    """
    parameters = ["A"]

    def __init__(self, domain=None, **kwargs):
        self._domain = domain
        super().__init__(**kwargs)

    @timedmethod
    @torch.compile
    def h(self, state):
        A = state.material[self.A]
        if self._domain != None:
            A = A * self._domain[:,:,:,None]
        Ms = state.material["Ms"]
        m = state.m
        dx = state.mesh.dx_tensor[0].reshape(-1,1,1,1)
        dy = state.mesh.dx_tensor[1].reshape(1,-1,1,1)
        dz = state.mesh.dx_tensor[2].reshape(1,1,-1,1)
        h = torch.zeros_like(state.m)

        # x
        if state.mesh.pbc[0] == 0:
            A_avg = 2.*A[1:,:,:]*A[:-1,:,:] / (A[1:,:,:]*dx[:-1,:,:,:] + A[:-1,:,:]*dx[1:,:,:,:])
            h[:-1,:,:,:] += A_avg * (m[ 1:,:,:,:]-m[:-1,:,:,:]) / dx[:-1,:,:,:] # m_i-1 - m_i
            h[ 1:,:,:,:] += A_avg * (m[:-1,:,:,:]-m[ 1:,:,:,:]) / dx[ 1:,:,:,:] # m_i+1 - m_i
        else:
            A_next = torch.roll(A, +1, dims=0)
            dx_next = torch.roll(dx, +1, dims=0)
            A_avg = 2.*A_next * A / (A_next*dx + A*dx_next)
            h += A_avg * (torch.roll(state.m, +1, dims=0) - state.m) / dx # m_i+1 - m_i

            A_avg = torch.roll(A_avg, -1, dims=0)
            h += A_avg * (torch.roll(state.m, -1, dims=0) - state.m) / dx # m_i-1 - m_i

        # y
        if state.mesh.pbc[1] == 0:
            A_avg = 2.*A[:,1:,:]*A[:,:-1,:] / (A[:,1:,:]*dy[:,:-1,:,:] + A[:,:-1,:]*dy[:,1:,:,:])
            h[:,:-1,:,:] += A_avg * (m[:, 1:,:,:]-m[:,:-1,:,:]) / dy[:,:-1,:,:] # m_i-1 - m_i
            h[:, 1:,:,:] += A_avg * (m[:,:-1,:,:]-m[:, 1:,:,:]) / dy[:, 1:,:,:] # m_i+1 - m_i
        else:
            A_next = torch.roll(A, +1, dims=1)
            dy_next = torch.roll(dy, +1, dims=1)
            A_avg = 2. * A_next * A / (A_next*dy + A*dy_next)
            h += A_avg * (torch.roll(state.m, +1, dims=1) - state.m) / dy # m_i+1 - m_i

            A_avg = torch.roll(A_avg, -1, dims=1)
            h += A_avg * (torch.roll(state.m, -1, dims=1) - state.m) / dy # m_i-1 - m_i

        # z
        if state.mesh.pbc[2] == 0:
            A_avg = 2.*A[:,:,1:]*A[:,:,:-1] / (A[:,:,1:]*dz[:,:,:-1,:] + A[:,:,:-1]*dz[:,:,1:,:])
            h[:,:,:-1,:] += A_avg * (m[:,:, 1:,:]-m[:,:,:-1,:]) / dz[:,:,:-1,:] # m_i-1 - m_i
            h[:,:, 1:,:] += A_avg * (m[:,:,:-1,:]-m[:,:, 1:,:]) / dz[:,:, 1:,:] # m_i+1 - m_i
        else:
            A_next = torch.roll(A, +1, dims=2)
            dz_next = torch.roll(dz, +1, dims=2)
            A_avg = 2. * A_next * A / (A_next*dz + A*dz_next)
            h += A_avg * (torch.roll(state.m, +1, dims=2) - state.m) / dz # m_i+1 - m_i

            A_avg = torch.roll(A_avg, -1, dims=2)
            h += A_avg * (torch.roll(state.m, -1, dims=2) - state.m) / dz # m_i-1 - m_i

        h *= 2. / (constants.mu_0 * Ms)
        return h.nan_to_num(posinf=0, neginf=0)



#
# This file is part of the magnum.np distribution
# (https://gitlab.com/magnum.np/magnum.np).
# Copyright (c) 2023 magnum.np team.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

from magnumnp.solvers.eigensolver import *
from magnumnp.solvers.llg import *
from magnumnp.solvers.ode_solvers import *
from magnumnp.solvers.ohm_solver import *
from magnumnp.solvers.string import *
from magnumnp.solvers.minimize import *
from magnumnp.solvers.steepest_descent import *
from magnumnp.solvers.LBFGS import *

__all__ = (eigensolver.__all__ +
           llg.__all__ +
           ode_solvers.__all__ +
           ohm_solver.__all__ +
           string.__all__ +
           minimize.__all__ +
           steepest_descent.__all__ +
           LBFGS.__all__)



#
# This file is part of the magnum.np distribution
# (https://gitlab.com/magnum.np/magnum.np).
# Copyright (c) 2023 magnum.np team.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

import torch
from magnumnp.common import logging
from scipy.integrate import ode

__all__ = ["ScipyODE"]

class ScipyODE(object):
    def __init__(self, f, name = "vode", method = "BDF", rtol = 1e-5, atol = 1e-5):
        self._f = f
        self._solver = ode(self._f_wrapper)
        self._solver.set_integrator(name = name,
                                    method = method,
                                    rtol = rtol,
                                    atol = atol)
        self._initialized = False
        logging.info_green("[LLGSolver] using Scipy ODE solver '%s' (method = '%s', rtol = %g, atol = %g)" % (name, method, rtol, atol))

    def _f_wrapper(self, t, x, kwargs): # TODO: should x be a 1D array?
        x_torch = torch.tensor(x.reshape(kwargs["state"].mesh.n + (-1,), order = "F"))
        f = self._f(t, x_torch, **kwargs)
        return f.detach().cpu().numpy().flatten(order = "F")

    def step(self, t, x_torch, dt, rtol = None, atol = None, **kwargs):
        if not self._initialized:
            x = x_torch.detach().cpu().numpy().reshape(-1, order = 'F')
            self._solver.set_initial_value(x, t)
            self._initialized = True

        self._solver.set_f_params(kwargs)

        t1 = self._solver.t + dt
        x1 = self._solver.integrate(t1)
        if not self._solver.successful():
            logging.warning("[LLGSolver] Scipy ODE solver: integration not successful!")

        return self._solver.t, torch.tensor(x1.reshape(x_torch.shape, order = "F"))



#
# This file is part of the magnum.np distribution
# (https://gitlab.com/magnum.np/magnum.np).
# Copyright (c) 2023 magnum.np team.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

from magnumnp.common import logging, timedmethod, constants, Timer, complex_dtype
from .field_terms import LinearFieldTerm
from .demag import demag_f, demag_g
import numpy as np
import torch
from time import time
import os

__all__ = ["DemagFieldNonEquidistant"]

class DemagFieldNonEquidistant(LinearFieldTerm):
    r"""
    Demagnetization Field:

    The dipole-dipole interaction gives rise to a long-range interaction.
    The integral formulation of the corresponding Maxwell equations can
    be represented as convolution of the magneti_dstation :math:`\vec{M} = M_s \; \vec{m}` with a proper
    demagneti_dstation kernel :math:`\vec{N}`

    .. math::
        \vec{h}^\text{dem}_{\vec{i}} = \sum\limits_{\vec{j}} \vec{N}_{\vec{i} - \vec{j}} \, \vec{M}_{\vec{j}},

    The convolution can be evaluated efficiently using an FFT method.

    :param p: number of next neighbors for near field via Newell's equation (default = 20)
    :type p: int, optional
    """
    def __init__(self, p = 20):
        self._p = p

    def _shape(self, state):
        s = [1,1,1]
        for i in range(2):
            if state.mesh.n[i] == 1:
                continue
            if state.mesh.pbc[i] == 0:
                s[i] = 2*state.mesh.n[i]
            else:
                s[i] = state.mesh.n[i] # no need to pad if nonzero pbc
        return s

    def _init_N_component(self, state, i_dst, i_src, perm, func):
        # rescale dx to avoid NaNs when using single precision
        # TODO: add scale to state and rescale like in DemagField
        dx = state.mesh.dx
        z = (torch.cumsum(state.mesh.dx_tensor[2], dim=0) - state.mesh.dx_tensor[2][0])
        dx_dst = torch.tensor([[dx[0], dx[1], state.mesh.dx_tensor[2][i_dst]][ind] for ind in perm])
        dx_src = torch.tensor([[dx[0], dx[1], state.mesh.dx_tensor[2][i_src]][ind] for ind in perm])

        shape = self._shape(state)
        ij = [torch.fft.fftfreq(n,1/n) for n in shape] # local indices
        ij[2] = ij[2]*0. + z[i_dst] - z[i_src] # use fixed distance for z-direction
        ij = torch.meshgrid(*ij,indexing='ij')
        x, y, z = [[ij[0]*dx[0], ij[1]*dx[1], ij[2].clone()][ind] for ind in perm]

        Lx = [state.mesh.n[0]*dx[0], state.mesh.n[1]*dx[1], torch.cumsum(state.mesh.dx_tensor[2], dim=0)[-1]]
        Lx = [Lx[ind] for ind in perm]

        offsets = [torch.arange(-state.mesh.pbc[ind], state.mesh.pbc[ind]+1) for ind in perm] # offset of pseudo PBC images
        offsets = torch.stack(torch.meshgrid(*offsets, indexing="ij"), dim=-1).flatten(end_dim=-2)

        Nc = torch.zeros(shape)
        for offset in offsets:
            Nc += func(x + offset[0]*Lx[0], y + offset[1]*Lx[1], z + offset[2]*Lx[2], *dx_dst, *dx_src, self._p)

        dim = [i for i in range(2) if state.mesh.n[i] > 1]
        if len(dim) > 0:
            Nc = torch.fft.rfftn(Nc, dim = dim) # seems to be complex!!
        return Nc # .real.clone()

    def _init_N(self, state):
        if isinstance(state.mesh.dx[2], float):
            logging.warning("mesh.dx[2] should not be constant when using DemagFieldNonEquidistant! Use the equidistant DemagField otherwise!")
        if not all([isinstance(dx, float) for dx in state.mesh.dx[:2]]):
            raise ValueError("Demag field only implemented for non-equidistant z-spacings. mesh.dx[0] and mesh.dx[1] need to be constant!")

        dtype = torch.get_default_dtype()
        torch.set_default_dtype(torch.float64) # always use double precision

        time_kernel = time()
        self._N = [None]*state.mesh.n[2]
        for i_dst in range(state.mesh.n[2]):
            self._N[i_dst] = [None]*state.mesh.n[2]
            for i_src in range(i_dst+1):
                Nxx = self._init_N_component(state, i_dst, i_src, [0,1,2], demag_f).to(dtype=complex_dtype[dtype])
                Nxy = self._init_N_component(state, i_dst, i_src, [0,1,2], demag_g).to(dtype=complex_dtype[dtype])
                Nxz = self._init_N_component(state, i_dst, i_src, [0,2,1], demag_g).to(dtype=complex_dtype[dtype])
                Nyy = self._init_N_component(state, i_dst, i_src, [1,2,0], demag_f).to(dtype=complex_dtype[dtype])
                Nyz = self._init_N_component(state, i_dst, i_src, [1,2,0], demag_g).to(dtype=complex_dtype[dtype])
                Nzz = self._init_N_component(state, i_dst, i_src, [2,0,1], demag_f).to(dtype=complex_dtype[dtype])

                self._N[i_dst][i_src] = [[ Nxx,  Nxy,  Nxz],
                                         [ Nxy,  Nyy,  Nyz],
                                         [ Nxz,  Nyz,  Nzz]]
                self._N[i_src][i_dst] = [[ Nxx,  Nxy, -Nxz],
                                         [ Nxy,  Nyy, -Nyz],
                                         [-Nxz, -Nyz,  Nzz]]
        logging.info(f"[DEMAG]: Time calculation of demag kernel = {time() - time_kernel} s")
        torch.set_default_dtype(dtype) # restore dtype

    @timedmethod
    def h(self, state):
        if not hasattr(self, "_N"):
            self._init_N(state)
        dim = [i for i in range(2) if state.mesh.n[i] > 1]
        shape = self._shape(state)
        s = [shape[i] for i in dim]

        if len(dim) == 0: # single spin   TODO: remove this when torch issue #96518 has been solved
            m_pad_fft = state.material["Ms"] * state.m
        else:
            m_pad_fft = torch.fft.rfftn(state.material["Ms"] * state.m, dim = dim, s = s)

        hx = torch.zeros(m_pad_fft[...,0].shape, dtype=self._N[0][0][0][0].dtype)
        hy = torch.zeros(m_pad_fft[...,0].shape, dtype=self._N[0][0][0][0].dtype)
        hz = torch.zeros(m_pad_fft[...,0].shape, dtype=self._N[0][0][0][0].dtype)

        for i_dst in range(state.mesh.n[2]):
            for i_src in range(state.mesh.n[2]):
                for ax in range(3):
                    hx[:,:,i_src] += self._N[i_src][i_dst][0][ax][:,:,0]*m_pad_fft[:,:,i_dst,ax]
                    hy[:,:,i_src] += self._N[i_src][i_dst][1][ax][:,:,0]*m_pad_fft[:,:,i_dst,ax]
                    hz[:,:,i_src] += self._N[i_src][i_dst][2][ax][:,:,0]*m_pad_fft[:,:,i_dst,ax]

        if len(dim) == 0: # single spin   TODO: remove this when torch issue #96518 has been solved
            hx = hx.real.clone()
            hy = hy.real.clone()
            hz = hz.real.clone()
        else:
            hx = torch.fft.irfftn(hx, dim = dim)
            hy = torch.fft.irfftn(hy, dim = dim)
            hz = torch.fft.irfftn(hz, dim = dim)

        return torch.stack([hx[:state.mesh.n[0],:state.mesh.n[1],:state.mesh.n[2]],
                            hy[:state.mesh.n[0],:state.mesh.n[1],:state.mesh.n[2]],
                            hz[:state.mesh.n[0],:state.mesh.n[1],:state.mesh.n[2]]], dim=3)



#
# This file is part of the magnum.np distribution
# (https://gitlab.com/magnum.np/magnum.np).
# Copyright (c) 2023 magnum.np team.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

import time
from collections import OrderedDict
from magnumnp.common import tabulate, logging
from functools import wraps

try:
    import resource
except:
    resource = None

__all__ = ["Timer", "timedmethod", "TimedOperator"]

class Timer(object):
    _timers    = OrderedDict()
    _start = None
    _current = ''
    _options = {
        'active': False,
        'skip': 0,
        'log_mem': False
    }

    def __init__(self, name):
        r"""
        Utility class that supports basic timing/profiling features. Each timer is
        initialized with a unique name. Each call to the body is timed and counted.
        Results can be printed with by calling :code:`print_report()`.

        *Example*
            .. code-block:: python

                Timer.enable()

                x = 0.0
                for i in range(10):
                    with Timer("Outer Function"):
                        for j in range(10):
                            with Timer("Inner Function"):
                                x += 1.0

                Timer.print_report()

            results in:

            .. code-block:: none

                ==================================================================
                TIMER REPORT
                ==================================================================
                Operation                     No of calls        Avg time [ms]        Total time [s]
                ----------------    -------------    ---------------    ----------------
                Outer Function                             10            0.0245333                 0.000245333
                    Inner Function                        100            0.000665188             6.65188e-05
                ----------------    -------------    ---------------    ----------------
                Total                                                                                                    0.00105381
                ==================================================================
        """
        if not Timer._options['active']: return

        self.t = 0.0
        self._fullname = Timer._current + '###' + name
        if self._fullname in self._timers:
            self._data= Timer._timers[self._fullname]
        else:
            self._data = {
                'name': name,
                'calls': 0,
                'total_time': 0.0,
                'memory': 0.0,
                'parent': Timer._current
            }
            Timer._timers[self._fullname] = self._data

    def __enter__(self):
        if not Timer._options['active']: return self

        Timer._current = self._fullname
        self._data['start'] = time.perf_counter()
        if Timer._options['log_mem']:
            self._data['mem_start'] = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss
        return self

    def __exit__(self, e_typ, e_val, trcbak):
        if all((e_typ, e_val, trcbak)):
            raise e_typ from e_val

        if not Timer._options['active']: return self

        if Timer._options['skip'] <= self._data['calls']:
            self.t = time.perf_counter() - self._data['start']
            self._data['total_time'] += self.t
            if Timer._options['log_mem']:
                self.mem = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss - self._data['mem_start']
                self._data['memory'] += self.mem
        self._data['calls'] += 1
        Timer._current = self._data['parent']
        return self

    @staticmethod
    def print_report():
        """
        Print the timing report.
        """
        def get_entries(parent = '', indent = ''):
            result = []

            for entry in filter(lambda x: x[1]['parent'] == parent, Timer._timers.items()):
                timed_calls = entry[1]['calls'] - Timer._options['skip']
                if timed_calls == 0:
                    t_avg = None
                    t_tot = None
                    if Timer._options['log_mem']:
                        mem = None
                else:
                    t_avg = (entry[1]['total_time'] / timed_calls) * 1000
                    t_tot = entry[1]['total_time']
                    if Timer._options['log_mem']:
                        mem = entry[1]['memory'] / 1024.

                if Timer._options['log_mem']:
                    result.append([indent + entry[1]['name'], entry[1]['calls'], t_avg, t_tot, mem])
                else:
                    result.append([indent + entry[1]['name'], entry[1]['calls'], t_avg, t_tot])
                result += get_entries(entry[0], indent + '    ')

            return result

        entries = get_entries()

        measured_time = sum([x[1]['total_time'] for x in Timer._timers.items() if x[1]['parent'] == ''])
        total_time = time.perf_counter() - Timer._start
        missing_time = total_time - measured_time

        if Timer._options['log_mem']:
            if Timer._start is not None:
                entries.append(['Total', None, None, total_time, None])
                entries.append(['Missing', None, None, missing_time, None])
            table = tabulate(entries, ["Operation", "No of calls", "Avg time [ms]", "Total time [s]", "Memory [MB]"])
        else:
            if Timer._start is not None:
                entries.append(['Total', None, None, total_time])
                entries.append(['Missing', None, None, missing_time])
            table = tabulate(entries, ["Operation", "No of calls", "Avg time [ms]", "Total time [s]"])

        # insert separator before total line
        lines = table.split('\n')
        lines.insert(len(lines) - 2, lines[1])
        table = "\n".join(lines)

        width = len(table.split("\n")[1])
        print("")
        print("=" * width)
        print("TIMER REPORT")
        print("=" * width)
        print(table)
        print("=" * width)
        print("")
        missing = missing_time / total_time
        if missing > 0.2:
            logging.warning("Too much time missing (%.0f%%). Add some Timers for more complete timing!" % (missing * 100.))

    @staticmethod
    def reset():
        """
        Reset all timers.
        """
        Timer._timers    = OrderedDict()
        Timer._current = ''
        Timer._options = {
            'active': False,
            'skip': 0,
            'log_mem': False
        }

    @staticmethod
    def configure(**kwargs):
        """
        Configure the timer.

        *Arguments*
            active :class:`bool`
                Flag the activates/deactived all timers.
            skip :class:`int`
                Skip the timing of the first x calls to every timer. The calls are
                still counted, but not considered in the average execution time.
            log_mem :class:`bool`
                Log accumulated memory consumption of timed calls
        """
        for (key, value) in kwargs.items():
            if key.lower() in Timer._options:
                Timer._options[key.lower()] = value
            else:
                raise ValueError("Option '%s' is not supported by Timer" % key)

        if Timer._options["log_mem"] == True and resource == None:
            raise RuntimeError("Import of 'resource' package failed. Try running Timer without 'log_mem = True'.")

    @staticmethod
    def disable():
        """
        Disable all timers.
        """
        Timer.configure(active = False)

    @staticmethod
    def enable(**kwargs):
        """
        Enable all timers. Takes all options that are accepted by :code:`configure`.
        """
        Timer._start = time.perf_counter()
        Timer.configure(active = True, **kwargs)


def timedmethod(method):
    @wraps(method)
    def timed(*args, **kwargs):
        with Timer("%s.%s" % (args[0].__class__.__name__, method.__name__)):
            return method(*args, **kwargs)
    return timed

def TimedOperator(obj, name="Unnamed Operator"):
    class wrapper(obj.__class__):
        def matvec(self, vec):
            with Timer(self._timer_name):
                return super(wrapper, self).matvec(vec)
    obj.__class__ = wrapper
    obj._timer_name = name
    return obj



#
# This file is part of the magnum.np distribution
# (https://gitlab.com/magnum.np/magnum.np).
# Copyright (c) 2023 magnum.np team.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

from magnumnp.common import logging, timedmethod, constants, normalize
from .ode_solvers import RKF45
from .llg import LLGSolver
import torch
from xitorch.interpolate import Interp1D

__all__ = ["StringSolver"]

class StringSolver(object):
    def __init__(self, fields, num_images = 20, fix_start = False, fix_end = False, omega = 0., interpolation = "cspline"):
        self._num_images = num_images
        self._fix_start = fix_start
        self._fix_end = fix_end
        self._omega = omega
        self._interpolation = interpolation
        self._fields = fields
        self._llg = LLGSolver(self._fields)


    def E(self, state, images):
        E = []
        for image in images:
            state.m = image
            E.append(sum([field.E(state) for field in self._fields]))
        return E


    def step(self, state, images, h = 1e-11, level = True):
        """
        Make one Step steepest descent along the gradient for each image.
        """
        for i, image in enumerate(images):
            if self._fix_start and i == 0: continue
            if self._fix_end and i == len(images): continue
            state.t = 0.0
            state.m = image
            llg = LLGSolver(self._fields, no_precession = True)
            llg.step(state, h)
            images[i] = state.m

        if level:
            newimages = []
            limages = self.level(state, images)
            for i in range(limages.shape[-1]):
                image = limages[:,:,:,:, i]
                newimages.append(image)
            images = newimages

        return images


    def level(self, state, images):
        logging.info("[StringSolver]: Leveling Images")
        Epath = self.E(state, images)
        source_data = torch.stack(images, dim = -1)
        Emin = min(Epath)
        Emax = max(Epath)

        x_source = [0.]
        for i in range(1, (len(images))):
            w = (((Epath[i] + Epath[i-1])/2. - Emin) / (Emax - Emin) + 1.)**self._omega
            dm = torch.linalg.norm(images[i] - images[i-1])
            x_source.append(w * dm + x_source[-1])
        x_source = torch.tensor(x_source)
        x_target = torch.linspace(x_source[0], x_source[-1], self._num_images)

        # start interpolation
        try:
            target_data = Interp1D(x_source, source_data, method = self._interpolation)(x_target)
        except:
            target_data = Interp1D(x_source, source_data, 'linear')(x_target)

        for i in range(self._num_images):
            state.m = target_data[:,:,:,:, i]
            normalize(state.m)
            target_data[:,:,:,:,i] = state.m

        return target_data



from magnumnp.common import logging, timedmethod, constants
import torch

__all__ = ["MinimizerBB"]

class MinimizerBB(object):
    def __init__(self, terms):
        """ This class implements the direct energy minimizing algorithm introduced in [Exl2014]_.

        .. note:: This feature is experimental.

        *Example*
          .. code:: python

            state = State(mesh)
            minimizer = MinimizerBB([ExchangeField()])
            minimizer.minimize(state)

        *Arguments*
            terms ([:class:`LLGTerm`])
                List of LLG contributions to be considered for energy minimization
        """
        self._terms = terms

    def E(self, state):
        return sum([term.E(state) for term in self._terms])

    def h(self, state):
        return sum([term.h(state) for term in self._terms])

    def dm(self, state, h):
        return torch.linalg.cross(state.m, torch.linalg.cross(state.m, h))

    def _midpoint(self, m, h, tau):
        """
        Solving the the semi-implicit midpoint scheme:
            m_i+1 = m_i + tau * (m_i + m_i+1)/2 x (m_i x Heff[m_i])

        see "Abert, 'Efficient Energyminimization in Finite-Difference Micromagnetics', 2014"
        see "Goldfarb, 'A Curvilinear Search Method for p-Harmonic Flows on Spheres', 2009"
        """
        mxh = torch.linalg.cross(m, h)
        mx, my, mz = m.unbind(-1)
        mxh_x, mxh_y, mxh_z = mxh.unbind(-1)

        N = 4 + tau*tau * (mxh_x*mxh_x + mxh_y*mxh_y + mxh_z*mxh_z)
        return torch.stack([(4*mx + 4*tau * (mxh_y*mz - mxh_z*my) + tau*tau*mx * (+ mxh_x*mxh_x - mxh_y*mxh_y - mxh_z*mxh_z) + 2*tau*tau*mxh_x * (mxh_y*my + mxh_z*mz)) / N,
                            (4*my + 4*tau * (mxh_z*mx - mxh_x*mz) + tau*tau*my * (- mxh_x*mxh_x + mxh_y*mxh_y - mxh_z*mxh_z) + 2*tau*tau*mxh_y * (mxh_z*mz + mxh_x*mx)) / N,
                            (4*mz + 4*tau * (mxh_x*my - mxh_y*mx) + tau*tau*mz * (- mxh_x*mxh_x - mxh_y*mxh_y + mxh_z*mxh_z) + 2*tau*tau*mxh_z * (mxh_x*mx + mxh_y*my)) / N], dim=-1)
        # mumax code
        #// m = 1 / (4 + τ²(m x H)²) [{4 - τ²(m x H)²} m - 4τ(m x m x H)]
        #// note: torque from LLNoPrecess has negative sign

    def _linesearch(self, state, m0, h0, dm0, tau):
        r = 0.5  # Reduction factor
        c = 0.5  # Sufficient decrease parameter
        m = -(constants.mu_0*state.material["Ms"]*state.mesh.cell_volumes*dm0*dm0).sum()
        t = -c*m
        E0 = self.E(state)

        for j in range(10):
            state.m = self._midpoint(m0, h0, tau)
            E = self.E(state)
            if E0 - E >= tau * t:
                break
            tau *= r
            logging.info_blue("[MinimizerBB] Linesearch: %d, E=%g" % (j, E))

    @timedmethod
    def minimize(self, state, maxiter = 2000, dm_tol = 1., tau_min = 1e-13, tau_max = 1e-5):
        tau = tau_min
        steps = 0
        dm_max = 1e18
        m0 = state.m.clone()
        h0 = self.h(state)
        dm0 = self.dm(state, h0)

        for i in range(maxiter):
            #self._linesearch(state, m0, h0, dm0, tau)
            state.m = self._midpoint(m0, h0, tau)
            h = self.h(state)

            # compute s^n-1 for step-size control
            m_diff = state.m - m0

            # compute y^n-1 for step-size control
            dm = torch.linalg.cross(state.m, torch.linalg.cross(state.m, h))
            dm_diff = dm - dm0

            # compute dm_max as convergence indicator
            dm_max = dm.abs().max()
            if dm_max < dm_tol:
                logging.info_green("[MinimizerBB] Successfully converged (iter=%d, dm_tol = %g)" % (i, dm_tol))
                return True

            # next stepsize (alternate tau1 and tau2)
            if (i % 2 == 0):
                tau = (m_diff*m_diff).sum() / (m_diff*dm_diff).sum()
            else:
                tau = (m_diff*dm_diff).sum() / (dm_diff*dm_diff).sum()
            tau = max(min(abs(tau), tau_max), tau_min) #* tau_sign

            logging.info_blue("[MinimizerBB] Step: %d, Tau: %.5g, dm_max: %.5g" % (i, tau, dm_max))

            m0 = state.m.clone()
            h0 = h.clone()
            dm0 = dm.clone()

        logging.warning("[MinimizerBB] Terminated after maxiter = %d (dm = %g, dm_tol = %g)" % (maxiter, dm_max, dm_tol))
        return False



#
# This file is part of the magnum.np distribution
# (https://gitlab.com/magnum.np/magnum.np).
# Copyright (c) 2023 magnum.np team.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

from magnumnp.common import logging
import torch

__all__ = ["Mesh"]

class Mesh(object):
    def __init__(self, n, dx, origin=(0,0,0), pbc=(0,0,0)):
        self.n = tuple(n)
        self.dx = tuple(dx)
        self.origin = tuple(origin)
        self.pbc = tuple(pbc)

        self.is_equidistant = all([isinstance(dx, (float, int)) for dx in dx])
        self.dx_tensor = [torch.tensor(dx, dtype=torch.get_default_dtype()).expand(n) for n, dx in zip(n, dx)]

        # compute cell_volumes
        if self.is_equidistant:
            self.cell_volumes = dx[0] * dx[1] * dx[2]
        else:
            dx, dy, dz = torch.meshgrid([torch.tensor(dx) for dx in dx], indexing = "ij") # use expand for equidistant dimentions
            self.cell_volumes = (dx*dy*dz).expand(self.n).unsqueeze(-1)

    def __str__(self):
        str_dx = ["%g" % dx if isinstance(dx, (int,float)) else "XX" for dx in self.dx]
        if self.pbc[0] != 0 or self.pbc[1] != 0 or self.pbc[2] != 0:
            str_pbc = ", pbc=[%d,%d,%d])" % self.pbc
        else:
            str_pbc = ")"
        return "%dx%dx%d (dx= %s x %s x %s%s" % (*self.n, *str_dx, str_pbc)

    def SpatialCoordinate(self):
        x = self.dx_tensor[0].cumsum(0) - self.dx_tensor[0]/2. + self.origin[0]
        y = self.dx_tensor[1].cumsum(0) - self.dx_tensor[1]/2. + self.origin[1]
        z = self.dx_tensor[2].cumsum(0) - self.dx_tensor[2]/2. + self.origin[2]

        XX, YY, ZZ = torch.meshgrid(x, y, z, indexing = "ij")
        return XX, YY, ZZ



#
# This file is part of the magnum.np distribution
# (https://gitlab.com/magnum.np/magnum.np).
# Copyright (c) 2023 magnum.np team.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

import torch
from ..common import normalize

__all__ = ["add_noise", "nsk", "hsl_to_rgb"]

def add_noise(x, dev = 1.0, mean = 0.0):
   if torch.is_tensor(x):
        x += torch.empty_like(x).normal_(mean = mean, std = dev)
        normalize(x)


def nsk(state): # TODO: document and improve interface
    m = state.m.mean(axis=2)
    dxm = torch.stack(torch.gradient(m, spacing = state.mesh.dx[0], dim = 0), dim = -1).squeeze(-1)
    dym = torch.stack(torch.gradient(m, spacing = state.mesh.dx[1], dim = 1), dim = -1).squeeze(-1)
    return 1./(4.*torch.pi) * (m * torch.linalg.cross(dxm, dym)).sum() * state.mesh.dx[0] * state.mesh.dx[1]


def hsl_to_rgb(h, s, l): # TODO: document and improve interface
    def hue_to_rgb(p, q, t):
        t = torch.where(t < 0, t + 1, t)
        t = torch.where(t > 1, t - 1, t)
        return torch.where(t < 1/6, p + (q - p) * 6 * t,
               torch.where(t < 1/2, q,
               torch.where(t < 2/3, p + (q - p) * (2/3 - t) * 6, p)))

    s = s.clamp(0, 1)
    l = l.clamp(0, 1)
    h = h.clamp(0, 1)

    q = l + s * torch.where(l < 0.5, l, 1 - l)
    p = 2 * l - q
    r = hue_to_rgb(p, q, h + 1/3)
    g = hue_to_rgb(p, q, h)
    b = hue_to_rgb(p, q, h - 1/3)

    return torch.stack([r, g, b], dim=-1)



#
# This file is part of the magnum.np distribution
# (https://gitlab.com/magnum.np/magnum.np).
# Copyright (c) 2023 magnum.np team.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

from magnumnp.utils.imaging_tools import *
from magnumnp.utils.logging_helpers import *
from magnumnp.utils.misc import *
from magnumnp.utils.voronoi import *

__all__ = (imaging_tools.__all__ +
           logging_helpers.__all__ +
           misc.__all__ +
           voronoi.__all__)



#
# This file is part of the magnum.np distribution
# (https://gitlab.com/magnum.np/magnum.np).
# Copyright (c) 2023 magnum.np team.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

from magnumnp.common import timedmethod, constants, logging
import torch
from .field_terms import FieldTerm, LinearFieldTerm
from torch import sin, cos

__all__ = ["UniaxialAnisotropyField", "CubicAnisotropyField", "CubicAnisotropyField2"]

class UniaxialAnisotropyField(LinearFieldTerm):
    r"""
    Uniaxial Anisotropy Field:

    .. math::

        \vec{h}^\text{u} = \frac{2 K_\text{u}}{\mu_0 \, M_s} \; \vec{e}_\text{u} \; (\vec{e}_\text{u} \cdot \vec{m}),

    with the anisotropy constant :math:`K_\text{u}` given in units of :math:`\text{J/m}^3`.

    :param Ku: Name of the material parameter for the anisotropy constant :math:`K_\text{u}`, defaults to "Ku"
    :type Ku: str, optional
    :param Ku_axis: Name of the material parameter for the anisotropy axis :math:`\vec{e}_\text{u}`, defaults to "Ku_axis"
    :type Ku_axis: str, optional
    """
    parameters = ["Ku", "Ku_axis"]

    @timedmethod
    @torch.compile
    def h(self, state):
        Ku = state.material[self.Ku]
        Ku_axis = state.material[self.Ku_axis]

        h = 2. * Ku * Ku_axis / (constants.mu_0 * state.material["Ms"]) * torch.sum(Ku_axis * state.m, dim=3, keepdim=True)
        return torch.nan_to_num(h, posinf=0, neginf=0)


class CubicAnisotropyField(FieldTerm):
    r"""
    Cubic Anisotropy Field:

    .. math::

        \vec{h}^\text{c} = -\frac{2 K_\text{c1}}{\mu_0 \, M_s} \; \begin{pmatrix} m_1 \, m_2^2 + m_1 \, m_3^2 \\ m_2 \, m_3^2 + m_2 \, m_1^2 \\ m_3 \, m_1^2 + m_3 \, m_2^2\end{pmatrix}
                           -\frac{2 K_\text{c2}}{\mu_0 \, M_s} \; \begin{pmatrix} m_1 \, m_2^2 \, m_3^2 \\ m_1^2 \, m_2 \, m_3^2 \\ m_1^2 \, m_2^2 \, m_3\end{pmatrix},

    with the anisotropy constants :math:`K_{c1}` and :math:`K_{c2}` given in units of :math:`\text{J/m}^3`.
    If Euler angles :math:`\alpha`, :math:`\beta` and :math:`\gamma` are provided, the cubic anisotropy axes are rotated such that the effective magnetization components :math:`m_i` with :math:`i \in \{x,y,z\}` read

    .. math::

      m_i = (\mat{A} \vec{e}_i) \cdot \vec{m}

    with

    .. math::

      \small
      \mat{A} = \begin{pmatrix}
            \cos(\alpha) \cos(\gamma) - \cos(\beta) \sin(\alpha) \sin(\gamma) & -\cos(\beta) \cos(\gamma) \sin(\alpha) - \cos(\alpha) \sin(\gamma) &  \sin(\alpha) \sin(\beta)\\
            \cos(\gamma) \sin(\alpha) + \cos(\alpha) \cos(\beta) \sin(\gamma) &  \cos(\alpha) \cos(\beta) \cos(\gamma) - \sin(\alpha) \sin(\gamma) & -\cos(\alpha) \sin(\beta)\\
            \sin(\beta) \sin(\gamma) & \cos(\gamma) \sin(\beta) & \cos(\beta)
            \end{pmatrix}

    :param Kc1: Name of the material parameter for the anisotropy constant :math:`K_\text{c1}`, defaults to "Kc1"
    :type Kc1: str, optional
    :param Kc2: Name of the material parameter for the anisotropy constant :math:`K_\text{c2}`, defaults to "Kc2"
    :type Kc2: str, optional
    :param Kc_alpha: Euler angle :math:`\alpha` for the rotation of the anisotropy axes
    :type Kc_alpha: str, optional
    :param Kc_beta: Euler angle :math:`\beta` for the rotation of the anisotropy axes
    :type Kc_beta: str, optional
    :param Kc_gamma: Euler angle :math:`\gamma` for the rotation of the anisotropy axes
    :type Kc_gamma: str, optional
    """
    parameters = ["Kc1", "Kc2", "Kc_alpha", "Kc_beta", "Kc_gamma"]
    def __init__(self, **kwargs):
        logging.warning("[CubicAnisotropyField] This field Term is deprecated due to bad performance (and will be removed in future versions). Use CubicAnisotropyField2() and provide an orthonormal set of axes instead of Euler angles.")
        super().__init__(**kwargs)

    def _R(self, state):
        a = state.material[self.Kc_alpha]
        b = state.material[self.Kc_beta]
        g = state.material[self.Kc_gamma]

        R = torch.stack([torch.concat([cos(a)*cos(g) - cos(b)*sin(a)*sin(g), -cos(b)*cos(g)*sin(a) - cos(a)*sin(g),  sin(a)*sin(b)], dim = -1),
                         torch.concat([cos(g)*sin(a) + cos(a)*cos(b)*sin(g),  cos(a)*cos(b)*cos(g) - sin(a)*sin(g), -cos(a)*sin(b)], dim = -1),
                         torch.concat([sin(b)*sin(g),                         cos(g)*sin(b),                         cos(b)       ], dim = -1)], dim = -1)
        return R

    @timedmethod
    @torch.compile
    def h(self, state):
        Kc1 = state.material[self.Kc1]
        Kc2 = state.material[self.Kc2]

        R = self._R(state)
        mx, my, mz = torch.einsum('...a, ...ab-> ...b', state.m, R).unbind(dim=-1) # matmult

        h =  2. * Kc1 * torch.stack([mx * (my**2 + mz**2), my*(mz**2 + mx**2), mz*(mx**2 + my**2)], dim = -1) + \
             2. * Kc2 * torch.stack([mx * my**2. * mz**2., mx**2. * my * mz**2., mx**2. * my**2. * mz], dim = -1)
        h = torch.einsum('...a, ...ba-> ...b', h, R) # matmult transpose
        return torch.nan_to_num(-1./constants.mu_0/state.material["Ms"] * h, posinf=0, neginf=0)

    @torch.compile
    def E(self, state):
        R = self._R(state)
        mx, my, mz = torch.einsum('...a, ...ab-> ...b', state.m, R).unbind(dim=-1) # matmult

        return ((state.material[self.Kc1] * (mx**2 * my**2 + mx**2 * mz**2 + my**2 * mz**2) +
                 state.material[self.Kc2] * (mx**2 * my**2 * mz**2)) * state.mesh.cell_volumes).sum()


class CubicAnisotropyField2(FieldTerm):
    r"""
    Cubic Anisotropy Field (alternative definition):

    .. math::

        \vec{h}^\text{c} = - \frac{2 K_\text{c1}}{\mu_0 \, M_s} \, (&((\vec{c_2} \cdot \vec{m})^2 + (\vec{c_3} \cdot \vec{m})^2) \; (\vec{c_1} \cdot \vec{m}) \; \vec{c_1} + \\
                                                                    &((\vec{c_1} \cdot \vec{m})^2 + (\vec{c_3} \cdot \vec{m})^2) \; (\vec{c_2} \cdot \vec{m}) \; \vec{c_2} + \\
                                                                    &((\vec{c_1} \cdot \vec{m})^2 + (\vec{c_2} \cdot \vec{m})^2) \; (\vec{c_3} \cdot \vec{m}) \; \vec{c_3})  \\
                           - \frac{2 K_\text{c2}}{\mu_0 \, M_s} \, (&(\vec{c_2} \cdot \vec{m})^2 \; (\vec{c_3} \cdot \vec{m})^2 \; (\vec{c_1} \cdot \vec{m}) \; \vec{c_1} +  \\
                                                                    &(\vec{c_1} \cdot \vec{m})^2 \; (\vec{c_3} \cdot \vec{m})^2 \; (\vec{c_2} \cdot \vec{m}) \; \vec{c_2} +  \\
                                                                    &(\vec{c_1} \cdot \vec{m})^2 \; (\vec{c_2} \cdot \vec{m})^2 \; (\vec{c_3} \cdot \vec{m}) \; \vec{c_3})   \\
                           - \frac{4 K_\text{c3}}{\mu_0 \, M_s} \, (&((\vec{c_2} \cdot \vec{m})^4 + (\vec{c_3} \cdot \vec{m})^4) \; (\vec{c_1} \cdot \vec{m})^3 \; \vec{c_1} +  \\
                                                                    &((\vec{c_1} \cdot \vec{m})^4 + (\vec{c_3} \cdot \vec{m})^4) \; (\vec{c_2} \cdot \vec{m})^3 \; \vec{c_2} +  \\
                                                                    &((\vec{c_1} \cdot \vec{m})^4 + (\vec{c_2} \cdot \vec{m})^4) \; (\vec{c_3} \cdot \vec{m})^3 \; \vec{c_3})

    with the anisotropy constants :math:`K_{c1}` and :math:`K_{c2}` given in units of :math:`\text{J/m}^3`.

    **NOTE:** Orthonormal axes need to be provided by the user!

    :param Kc1: Name of the material parameter for the anisotropy constant :math:`K_\text{c1}`, defaults to "Kc1"
    :type Kc1: str, optional
    :param Kc2: Name of the material parameter for the anisotropy constant :math:`K_\text{c2}`, defaults to "Kc2"
    :type Kc2: str, optional
    :param Kc_axis1: Name of the material parameter for the 1. axis
    :type Kc_axis1: str, optional
    :param Kc_axis2: Name of the material parameter for the 2. axis
    :type Kc_axis2: str, optional
    """
    parameters = ["Kc1", "Kc2", "Kc3", "Kc_axis1", "Kc_axis2"]

    @timedmethod
    @torch.compile
    def h(self, state):
        Kc1 = state.material[self.Kc1]
        Kc2 = state.material[self.Kc2]
        Kc3 = state.material[self.Kc3]

        Kc_axis1 = state.material[self.Kc_axis1]
        Kc_axis2 = state.material[self.Kc_axis2]
        Kc_axis3 = torch.linalg.cross(Kc_axis1, Kc_axis2)

        c1m = (Kc_axis1 * state.m).sum(axis=-1, keepdims=True)
        c2m = (Kc_axis2 * state.m).sum(axis=-1, keepdims=True)
        c3m = (Kc_axis3 * state.m).sum(axis=-1, keepdims=True)

        h  = 2 * Kc1 * ((c2m**2 + c3m**2) * c1m * Kc_axis1 +    \
                        (c1m**2 + c3m**2) * c2m * Kc_axis2 +    \
                        (c1m**2 + c2m**2) * c3m * Kc_axis3) +   \
             2 * Kc2 * (c2m**2 * c3m**2 * c1m * Kc_axis1 +      \
                        c1m**2 * c3m**2 * c2m * Kc_axis2 +      \
                        c1m**2 * c2m**2 * c3m * Kc_axis3) +     \
             4 * Kc3 * ((c2m**4 + c3m**4) * c1m**3 * Kc_axis1 + \
                        (c1m**4 + c3m**4) * c2m**3 * Kc_axis2 + \
                        (c1m**4 + c2m**4) * c3m**3 * Kc_axis3)

        return torch.nan_to_num(-1./constants.mu_0/state.material["Ms"] * h, posinf=0, neginf=0)

    @torch.compile
    def E(self, state):
        Kc1 = state.material[self.Kc1]
        Kc2 = state.material[self.Kc2]
        Kc3 = state.material[self.Kc3]

        Kc_axis1 = state.material[self.Kc_axis1]
        Kc_axis2 = state.material[self.Kc_axis2]
        Kc_axis3 = torch.linalg.cross(Kc_axis1, Kc_axis2)

        c1m = (Kc_axis1 * state.m).sum(axis=-1, keepdims=True)
        c2m = (Kc_axis2 * state.m).sum(axis=-1, keepdims=True)
        c3m = (Kc_axis3 * state.m).sum(axis=-1, keepdims=True)

        E = Kc1 * ((c1m**2 * c2m**2) +       \
                   (c1m**2 * c3m**2) +       \
                   (c2m**2 * c3m**2)) +      \
            Kc2 * c1m**2 * c2m**2 * c3m**2 + \
            Kc3 * ((c1m**4 * c2m**4) +       \
                   (c1m**4 * c3m**4) +       \
                   (c2m**4 * c3m**4))
        return (E * state.mesh.cell_volumes).sum()



#
# This file is part of the magnum.np distribution
# (https://gitlab.com/magnum.np/magnum.np).
# Copyright (c) 2023 magnum.np team.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

import torch
from magnumnp.common import logging

__all__ = ["RKF45"]

#Runge-Kutta-Fehlberg method with stepsize control
class RKF45(object):
    def __init__(self, f, dt = 1e-15, atol = 1e-5, rtol = 1e-5):
        self._f = f
        self._dt = dt
        self._order = 4

        # Numerical Recipies 3rd Edition suggests these values:
        self._headroom = 0.9
        self._maxstep = 1e-10
        self._minscale = 0.2
        self._maxscale = 10.
        self._atol = atol
        self._rtol = rtol # currently rtol is ignored since m is normalized!
        logging.info_green("[LLGSolver] using RKF45 solver (atol = %g)" % atol)

    def _try_step(self, t, x, **kwargs):
        f, dt = self._f, self._dt
        kwargs["state"]._dt = dt  # update current dt in state used by thermal field class
        k1 = dt * f(t,              x, **kwargs)
        k2 = dt * f(t +  1./ 4.*dt, x +      1./ 4.*k1, **kwargs)
        k3 = dt * f(t +  3./ 8.*dt, x +      3./32.*k1 +      9./32.*k2, **kwargs)
        k4 = dt * f(t + 12./13.*dt, x + 1932./2197.*k1 - 7200./2197.*k2 + 7296./2197.*k3, **kwargs)
        k5 = dt * f(t +      1.*dt, x +   439./216.*k1 -          8.*k2 + 3680./ 513.*k3 -  845./4104.*k4, **kwargs)
        k6 = dt * f(t +  1./ 2.*dt, x -    8. / 27.*k1 +          2.*k2 - 3544./2565.*k3 + 1859./4104.*k4 - 11./40.*k5, **kwargs)

        dx = 16./135.*k1 + 6656./12825.*k3 + 28561./56430.*k4 - 9./50.*k5 + 2./55.*k6
        rk_error = dx - (25./216.*k1 + 1408./2565.*k3 + 2197./4104.*k4 - 1./5.*k5)
        return x+dx, t+dt, rk_error

    def _optimal_stepsize(self, rk_error, atol):
        norm = torch.linalg.norm(rk_error.flatten() / atol, torch.inf)
        if torch.isnan(norm):
            raise RuntimeError("Unexpected error norm= %.5g!" % norm)

        if norm > 1.1:
            # decrease step, no more than factor of 5, but a fraction S more
            # than scaling suggests (for better accuracy)
            r = self._headroom / torch.pow(norm, 1.0/self._order)
            if (r < self._minscale):
                r = self._minscale
        elif norm < 0.5:
            # increase step, but no more than by a factor of 5
            r = self._headroom / torch.pow(norm, 1.0/(self._order+1.0));
            if r > self._maxscale: # increase no more than factor of 5
                r = self._maxscale
            if r < 1.: # don't allow any decrease caused by S<1
                r = 1.
        else: # no change
            r = 1.

        dt_opt = self._dt * r
        if dt_opt > self._maxstep:
            dt_opt = self._maxstep
        return float(dt_opt)

    def step(self, t, x, dt, rtol = None, atol = None, **kwargs):
        t1 = t + dt
        while t < t1:
            _x1, _t1, err = self._try_step(t, x, **kwargs)
            dt_opt = self._optimal_stepsize(err, atol or self._atol)
            if self._dt > dt_opt or self._dt > t1 - t:
                # step size was too large, retry with optimal stepsize
                self._dt = min(dt_opt, t1 - t)
                logging.debug("REVERT step: %g, new step size: %g, time: %g" % (self._dt, dt_opt, t))
            else:
                # accept step, adapt stepsize for next step
                x = _x1
                t = _t1
                logging.debug("ACCEPT step: %g, new step size: %g, time: %g" % (self._dt, dt_opt, t))
                self._dt = dt_opt
                kwargs["state"]._step += 1
        return t, x



#
# This file is part of the magnum.np distribution
# (https://gitlab.com/magnum.np/magnum.np).
# Copyright (c) 2023 magnum.np team.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

from magnumnp.field_terms.anisotropy import *
from magnumnp.field_terms.demag import *
from magnumnp.field_terms.demag_nonequidistant import *
from magnumnp.field_terms.demagPBC import *
from magnumnp.field_terms.dmi import *
from magnumnp.field_terms.exchange import *
from magnumnp.field_terms.external import *
from magnumnp.field_terms.field_terms import *
from magnumnp.field_terms.oersted import *
from magnumnp.field_terms.rkky import *
from magnumnp.field_terms.rux import *
from magnumnp.field_terms.spintorque import *
from magnumnp.field_terms.thermal import *
from magnumnp.field_terms.vector_potential import *

__all__ = (anisotropy.__all__ +
           demag.__all__ +
           demag_nonequidistant.__all__ +
           demagPBC.__all__ +
           dmi.__all__ +
           exchange.__all__ +
           external.__all__ +
           field_terms.__all__ +
           oersted.__all__ +
           rux.__all__ +
           rkky.__all__ +
           spintorque.__all__ +
           thermal.__all__ +
           vector_potential.__all__)



"""magnum.np main module"""

__version__ = '2.0.2'

import magnumnp.common.logging as logging
from magnumnp.common.utils import get_gpu_with_least_memory
import torch
import os

# set default device and dtype
torch.set_default_dtype(torch.float64)
device_id = os.environ.get('CUDA_DEVICE')
if device_id == None:
    device_id = get_gpu_with_least_memory()
device = torch.device(f"cuda:{device_id}" if int(device_id) >= 0 else "cpu")
torch.set_default_device(device)
torch.manual_seed(2147483647) # fix seed

# monkey patch torch versions < 2.0 or on Windows
try:
   def fake_compile(func):
       return func

   @torch.compile
   def dummy_function():
      return
   dummy_function()
except Exception as e:
    torch.compile = fake_compile
    logging.warning(e)


try:
    import setproctitle
    setproctitle.setproctitle("magnumnp")

    from magnumnp.common import *
    from magnumnp.field_terms import *
    from magnumnp.solvers import *
    from magnumnp.loggers import *
    from magnumnp.utils import *

    logging.info_green("magnum.np %s" % __version__)

except Exception as e:
    import magnumnp.common.logging as logging
    logging.error(str(e).split("\n")[0])
    for line in str(e).split("\n")[1:]:
        logging.info(line)

    try:
        # do nothing if in IPython
        __IPYTHON__
        pass

    except NameError:
        # exit otherwise
        exit()



#
# This file is part of the magnum.np distribution
# (https://gitlab.com/magnum.np/magnum.np).
# Copyright (c) 2023 magnum.np team.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

from magnumnp.common.constants import *
from magnumnp.common.logging import *
from magnumnp.common.material import *
from magnumnp.common.mesh import *
from magnumnp.common.state import *
from magnumnp.common.tabulate import *
from magnumnp.common.timer import *
from magnumnp.common.io import *
from magnumnp.common.time_interpolator import *
from magnumnp.common.utils import *

__all__ = (["constants"] +
           logging.__all__ +
           material.__all__ +
           mesh.__all__ +
           state.__all__ +
           timer.__all__ +
           io.__all__ +
           time_interpolator.__all__ +
           utils.__all__)



#
# This file is part of the magnum.np distribution
# (https://gitlab.com/magnum.np/magnum.np).
# Copyright (c) 2023 magnum.np team.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

import torch
from magnumnp.common import logging
from scipy.integrate import odeint

__all__ = ["ScipyOdeint"]

class ScipyOdeint(object):
    def __init__(self, f, hmax = 0.0, hmin = 0.0, mxordn = 12, mxords = 5, rtol = 1e-4, atol = 1e-4):
        self._f = f
        self._hmax = hmax
        self._hmin = hmin
        self._mxordn = mxordn
        self._mxords = mxords
        self._rtol = rtol
        self._atol = atol

        logging.info_green("[LLGSolver] using Scipy odeint Solver (rtol = %g, atol = %g)" % (rtol, atol))

    def _f_wrapper(self, t, x, kwargs):
        x_torch = torch.tensor(x.reshape(kwargs["state"].mesh.n + (-1,), order = "F"))
        f = self._f(t * 1e-9, x_torch, **kwargs) * 1e-9 # scale time by 1e9 to prevent underflow error
        return f.detach().cpu().numpy().flatten(order = "F")

    def step(self, t, x_torch, dt, rtol = None, atol = None, **kwargs):
        x0 = x_torch.detach().cpu().numpy().reshape(-1, order = 'F')
        t1 = t + dt
        x1 = odeint(self._f_wrapper,
                    x0,
                    [(t*1e9).detach().cpu(), (t1*1e9).detach().cpu()],
                    args = (kwargs,),
                    rtol = rtol or self._rtol,
                    atol = atol or self._atol,
                    tfirst = True)[1]

        return t1, torch.tensor(x1.reshape(x_torch.shape, order = "F"))

    def solve(self, tt, x_torch, rtol = None, atol = None, **kwargs):
        x0 = x_torch.detach().cpu().numpy().reshape(-1, order = 'F')
        res = odeint(self._f_wrapper,
                     x0,
                     tt.detach().cpu().numpy()*1e9,
                     args = (kwargs,),
                     rtol = rtol or self._rtol,
                     atol = atol or self._atol,
                     tfirst = True)

        return torch.tensor(res.reshape(tt.shape + x_torch.shape, order = "F"))



#
# This file is part of the magnum.np distribution
# (https://gitlab.com/magnum.np/magnum.np).
# Copyright (c) 2023 magnum.np team.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

from magnumnp.common import constants
import torch

__all__ = ["FieldTerm", "LinearFieldTerm"]

class FieldTerm(object):
    parameters = []

    def __init__(self, **kwargs):
        unknown_params = set(kwargs.keys()) - set(self.parameters)
        if unknown_params:
            raise Warning("Got unknown parameters '%s'. Ignoring!" % unknown_params)
        params = {key:key for key in self.parameters}
        params.update(kwargs)
        for key, value in params.items():
            setattr(self, key, value)

class LinearFieldTerm(FieldTerm):
    @torch.compile
    def E(self, state, domain = Ellipsis):
        E = -0.5 * constants.mu_0 * state.material["Ms"] * state.m * self.h(state) * state.mesh.cell_volumes
        return E[domain].sum()



#
# This file is part of the magnum.np distribution
# (https://gitlab.com/magnum.np/magnum.np).
# Copyright (c) 2023 magnum.np team.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

from magnumnp.common import timedmethod, constants
import torch

__all__ = ["SpinOrbitTorque", "SpinTorqueZhangLi", "SpinTorqueSlonczewski"]

#TODO: generalize interface (ZhangLi uses state.j, SOT uses material["je"]
class SpinOrbitTorque(object):
    r"""
    General spin torque contributions can be described by the following field

    .. math::
        \vec{h}^\text{sot} = -\frac{j_e \hbar}{2 e \mu_0 M_s} \left[\eta_\text{damp} \, \vec{m} \times \vec{p} + \eta_\text{field} \, \vec{p} \right],

    with the current density :math:`j_e`, the reduced Planck constant :math:`\hbar`,
    the elementary charge :math:`e`, and the polarization of the electrons :math:`\vec{p}`.
    :math:`\eta_\text{damp}` and :math:`\eta_\text{field}` are material parameters which
    describe the amplitude of damping- and field-like torque.

    In case of Spin-Orbit-Torqe (SOT) :math:`\eta_\text{field}` and :math:`\eta_\text{damp}` are constant material parameters.
    """
    @timedmethod
    @torch.compile
    def h(self, state):
        p = state.material["p"].expand_as(state.m)
        h = state.material["eta_damp"] * torch.linalg.cross(state.m, p) + state.material["eta_field"] * p
        h *= -state.material["je"] * constants.hbar / (2. * constants.e * state.material["Ms"] * constants.mu_0 * state.material["d"])
        return h.nan_to_num(posinf=0, neginf=0)

    def E(self, state):
        raise NotImplemented()


class SpinTorqueZhangLi(object):
    r"""
    Zhang Lie spin torque contributions can be described by the following field

    .. math::
        \vec{h}^\text{stt,zl} = \frac{b}{\gamma} \left[\vec{m} \times (\vec{j}_e \cdot \nabla) \vec{m} + \xi \; (\vec{j}_e \cdot \nabla) \vec{m} \right],

    with the reduced gyromagnetic ratio :math:`\gamma`, the degree of nonadiabacity :math:`\xi`. :math:`b` is the polarization rate of the conducting electrons and can be written as

    .. math::
        b = \frac{\beta \mu_B}{e M_s (1+\xi^2)},

    with the Bohr magneton :math:`\mu_B`, and the dimensionless polarization rate :math:`\beta`.
    """
    @timedmethod
    @torch.compile
    def h(self, state):
        dim = [i for i in range(3) if state.mesh.n[i] > 1]
        dx = [state.mesh.dx[i] for i in range(3) if state.mesh.n[i] > 1]

        j = state.j # (state.t) TODO: allow time-dependent j
        jgradm = torch.einsum('...a,...ba-> ...b', j[...,dim], torch.stack(torch.gradient(state.m, spacing=dx, dim=dim), dim=-1)) # matmult

        return state.material["b"] / constants.gamma * (torch.linalg.cross(state.m, jgradm) + state.material["xi"] * jgradm)

    def E(self, state):
        raise NotImplemented()


class SpinTorqueSlonczewski(object):
    r"""
    Slonczewski spin torque contributions can be described by the following field:

    .. math::
        \vec{h}^\text{stt,slonczewski} = \beta \left[ \epsilon (\vec{m} \times \vec{p}_\text{p}) + \epsilon' \vec{p}_\text{p} \right],

    with the polarization vector of the conducting electrons :math:`\vec{p}_\text{p}`
    and the secondary spin torque efficiency :math:`\epsilon'`.

    The :math:`\beta` is given by current density :math:`J`,
    the thickness of the fixed layer :math:`d` (default using the thickness of the mesh),
    the reduced Planck constant :math:`\hbar`,
    the elementary charge :math:`e`,
    the permeability of free space :math:`\mu_0`,

    .. math::
        \beta = \frac{\hbar J} {e M_s d \mu_0},

    The :math:`\epsilon` is computed by assuming both the fixed layer and free layer have the same spin polarization :math:`P` and the spin diffusion length :math:`\Lambda`: across the spacer layer

    .. math::
        \epsilon = \frac{P \Lambda^2} {(\Lambda^2 + 1) + (\Lambda^2 - 1) \vec{m} \cdot \vec{p}_\text{p}}.
    """
    @timedmethod
    @torch.compile
    def h(self, state):
        mp = state.material["mp"]
        Lambda = state.material["Lambda"]

        # use thickness of mesh if not provided
        if state.material["d"] is None:
            d = state.mesh.dx[2]
        else:
            d = state.material["d"]

        epsilon = state.material["P"] * Lambda**2 / ((Lambda**2 + 1) + ((Lambda**2 - 1) * (state.m*mp).sum(axis = 3, keepdim=True)))
        mxp = torch.linalg.cross(state.m, mp)
        h = epsilon * mxp + state.material["epsilon_prime"] * mp

        h *= constants.hbar * state.material["J"] / (constants.mu_0 * state.material["Ms"] * constants.e * d)
        return h.nan_to_num(posinf=0, neginf=0)

    def E(self, state):
        raise NotImplemented()



#!/usr/bin/env python
#
# This file is part of the magnum.np distribution
# (https://gitlab.com/magnum.np/magnum.np).
# Copyright (c) 2023 magnum.np team.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

from setuptools import setup

# read the contents of your README file
from pathlib import Path
this_directory = Path(__file__).parent
long_description = (this_directory / "README.rst").read_text()

setup(name='magnumnp',
      version='v2.0.2',
      description='magnum.np finite-difference package for the solution of micromagnetic problems',
      long_description=long_description,
      long_description_content_type='text/x-rst',
      author='Florian Bruckner',
      author_email='florian.bruckner@univie.ac.at',
      url='http://gitlab.com/magnum.np/magnum.np',
      project_urls = {'Documentation': 'https://magnum.np.gitlab.io/magnum.np/',
                      'Changelog': 'https://gitlab.com/magnum.np/magnum.np/blob/main/CHANGELOG'
      },
      packages=['magnumnp', 'magnumnp.common', 'magnumnp.field_terms', 'magnumnp.loggers', 'magnumnp.solvers', 'magnumnp.solvers.ode_solvers', 'magnumnp.utils'],
      install_requires = [
            'numpy',
            'pynvml',
            'pyvista',
            'scipy',
            'setproctitle',
            'torch',
            'torchdiffeq',
            'xitorch',
            'tqdm',
            ]
     )
